import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
import ta
import matplotlib.pyplot as plt

plt.rcParams['font.sans-serif']=['SimHei'] #用来正常显示中文标签
plt.rcParams['axes.unicode_minus'] = False#用来正常显示负号

def load_and_prepare_data(file_path):
    """
    加载并预处理Excel数据
    """
    # 读取Excel文件
    df = pd.read_excel(file_path)
    
    return df

def prepare_stock_data(df):
    """
    准备股票数据，计算技术指标
    """
    # 确保日期格式正确
    df['date'] = pd.to_datetime(df['date'])
    df = df.sort_values('date')
    
    # 基础价格特征
    df['price_range'] = df['high'] - df['low']  # 当日价格范围
    df['price_change'] = df['close'] - df['open']  # 当日价格变化
    df['price_return'] = df['close'].pct_change()  # 日收益率
    
    # 移动平均线
    df['MA5'] = df['close'].rolling(window=5).mean()
    df['MA10'] = df['close'].rolling(window=10).mean()
    df['MA20'] = df['close'].rolling(window=20).mean()
    df['MA60'] = df['close'].rolling(window=60).mean()
    
    # 移动平均线差值
    df['MA5_diff'] = df['close'] - df['MA5']
    df['MA10_diff'] = df['close'] - df['MA10']
    df['MA20_diff'] = df['close'] - df['MA20']
    
    # 计算成交量指标
    df['volume_MA5'] = df['volume'].rolling(window=5).mean()
    df['volume_MA10'] = df['volume'].rolling(window=10).mean()
    df['volume_ratio'] = df['volume'] / df['volume_MA5']
    
    # 波动率指标
    df['volatility_5'] = df['close'].rolling(window=5).std()
    df['volatility_10'] = df['close'].rolling(window=10).std()
    
    try:
        # RSI
        df['RSI'] = ta.momentum.rsi(df['close'], window=14)
        df['RSI_5'] = ta.momentum.rsi(df['close'], window=5)
        
        # MACD
        df['MACD'] = ta.trend.macd_diff(df['close'])
        df['MACD_signal'] = ta.trend.macd_signal(df['close'])
        
        # 布林带
        df['BB_high'] = ta.volatility.bollinger_hband(df['close'])
        df['BB_low'] = ta.volatility.bollinger_lband(df['close'])
        df['BB_mid'] = ta.volatility.bollinger_mavg(df['close'])
        df['BB_width'] = (df['BB_high'] - df['BB_low']) / df['BB_mid']
        
        # KDJ指标
        df['K'] = ta.momentum.stoch(df['high'], df['low'], df['close'])
        df['D'] = ta.momentum.stoch_signal(df['high'], df['low'], df['close'])
        
    except Exception as e:
        print(f"计算技术指标时出现错误: {e}")
        print("继续使用基本指标...")
    
    # 历史趋势特征
    df['prev_close'] = df['close'].shift(1)
    df['prev_volume'] = df['volume'].shift(1)
    df['close_5d_mean'] = df['close'].rolling(window=5).mean()
    df['close_5d_std'] = df['close'].rolling(window=5).std()
    
    # 删除NaN值
    df = df.dropna()
    
    return df

def create_features_targets(df, prediction_days=1):
    """
    创建特征和目标变量
    """
    # 创建目标变量 - 下一个交易日的收盘价
    df['target'] = df['close'].shift(-prediction_days)
    
    # 选择特征
    features = [
        'open', 'high', 'low', 'close', 'volume',
        'price_range', 'price_change', 'price_return',
        'MA5', 'MA10', 'MA20', 'MA60',
        'MA5_diff', 'MA10_diff', 'MA20_diff',
        'volume_MA5', 'volume_MA10', 'volume_ratio',
        'volatility_5', 'volatility_10',
        'prev_close', 'prev_volume',
        'close_5d_mean', 'close_5d_std'
    ]
    
    # 添加技术指标（如果存在）
    optional_features = ['RSI', 'RSI_5', 'MACD', 'MACD_signal', 
                        'BB_high', 'BB_low', 'BB_width', 'K', 'D']
    
    for feature in optional_features:
        if feature in df.columns:
            features.append(feature)
    
    # 删除有缺失值的行
    All = df[features]
    df = df.dropna()

    X = df[features]
    y = df['target']
    
    return All, X, y, features

def train_stock_model(processed_df, X, y, test_size=0.2):
    """
    训练随机森林模型
    """
    # 分割训练集和测试集
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, shuffle=False)
    
    # 标准化特征
    scaler = StandardScaler()
    X_train_scaled = scaler.fit_transform(X_train)
    X_test_scaled = scaler.transform(X_test)
    
    # 训练模型
    model = RandomForestRegressor(
        n_estimators=200,
        max_depth=20,
        min_samples_split=5,
        min_samples_leaf=2,
        random_state=42
    )
    model.fit(X_train_scaled, y_train)
    
    # 预测
    y_pred = model.predict(X_test_scaled)
    
    # 评估模型
    mse = mean_squared_error(y_test, y_pred)
    mae = mean_absolute_error(y_test, y_pred)
    r2 = r2_score(y_test, y_pred)
    
    print(f'均方误差 (MSE): {mse:.4f}')
    print(f'平均绝对误差 (MAE): {mae:.4f}')
    print(f'R2 分数: {r2:.4f}')
    
    # 计算预测准确率（方向）
    direction_accuracy = np.mean(
        (y_test.values[1:] - y_test.values[:-1]) * 
        (y_pred[1:] - y_pred[:-1]) > 0
    ) * 100
    print(f'涨跌方向准确率: {direction_accuracy:.2f}%')
    
    # 绘制预测结果
    plt.figure(figsize=(15, 6))
    
    # 绘制最近30天的预测结果
    plt.subplot(1, 2, 1)
    plt.plot(processed_df['date'][-30:], y_test.values[-30:], label='实际值', color='blue')
    plt.plot(processed_df['date'][-30:], y_pred[-30:], label='预测值', color='red', linestyle='--')
    plt.title('收盘价预测对比 (最近30天)')
    plt.xlabel('时间')
    plt.ylabel('价格')
    plt.legend()
    plt.xticks(rotation=45)
    
    # 绘制散点图
    plt.subplot(1, 2, 2)
    plt.scatter(y_test, y_pred, alpha=0.5)
    plt.plot([y_test.min(), y_test.max()], [y_test.min(), y_test.max()], 'r--', lw=2)
    plt.xlabel('实际值')
    plt.ylabel('预测值')
    plt.title('预测值 vs 实际值')
    
    plt.tight_layout()
    plt.show()
    
    return model, scaler, (X_test, y_test, y_pred)

def predict_next_day(model, scaler, latest_data):
    """
    预测下一个交易日的收盘价
    """
    latest_data_scaled = scaler.transform(latest_data)
    prediction = model.predict(latest_data_scaled)
    return prediction[0]

def analyze_feature_importance(model, feature_names):
    """
    分析特征重要性
    """
    importances = pd.DataFrame({
        'feature': feature_names,
        'importance': model.feature_importances_
    })
    importances = importances.sort_values('importance', ascending=False)
    
    plt.figure(figsize=(12, 6))
    plt.bar(range(len(importances.head(15))), importances['importance'].head(15))
    plt.xticks(range(len(importances.head(15))), importances['feature'].head(15), rotation=45, ha='right')
    plt.title('Top 15 特征重要性分析')
    plt.tight_layout()
    plt.show()
    
    return importances

# 主程序
if __name__ == "__main__":
    try:
        # 读取数据
        file_path = '上证指数历史数据.xlsx'
        print("正在读取数据...")
        raw_df = load_and_prepare_data(file_path)
        
        # 准备数据
        print("正在处理数据...")
        processed_df = prepare_stock_data(raw_df)
        
        # 创建特征和目标
        print("正在创建特征...")
        All, X, y, features = create_features_targets(processed_df)
        
        # 训练模型
        print("正在训练模型...")
        model, scaler, (X_test, y_test, y_pred) = train_stock_model(processed_df, X, y)
        
        # 分析特征重要性
        print("正在分析特征重要性...")
        feature_importance = analyze_feature_importance(model, features)
        
        # 预测下一个交易日
        latest_data = All.iloc[-1:] # 获取最新的一行数据

        next_day_prediction = predict_next_day(model, scaler, latest_data)
        current_close = latest_data['close'].iloc[0]
        
        print(f'\n当前收盘价: {current_close:.2f}')
        print(f'下一个交易日预测收盘价: {next_day_prediction:.2f}')
        print(f'预测变化: {(next_day_prediction - current_close):.2f} ({((next_day_prediction - current_close)/current_close*100):.2f}%)')
        
        # 打印最重要的特征
        print('\n最重要的10个特征:')
        print(feature_importance.head(10))
        
    except Exception as e:
        print(f"程序运行出错: {e}")
        raise e